"""Implements the LabelledDataGroup, which allows to handle groups and
containers that can be associated with further coordinates.

This imitates the xarray selection interface and provides a uniform interface
to select data from these groups. Most importantly, it allows to combine all
the data of one group, allowing to conveniently work with heterogeneously
stored data.
"""

import logging
from itertools import product
from typing import TypeVar, Tuple, Dict, Sequence, Union, List

import numpy as np
import xarray as xr

from . import OrderedDataGroup
from ..abc import AbstractDataContainer
from ..containers import XrDataContainer
from ..utils import extract_coords
from ..utils.coords import TCoord, TCoordsDict, TDims
from ..tools import apply_along_axis

# Local constants
log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------

class LabelledDataGroup(OrderedDataGroup):
    """A group that assumes that the members it contains can be labelled
    with dimension names and coordinates.

    Such a group has the great benefit to provide a selection interface that
    works fully on the dimension labels and coordinates and can cooperate with
    the xarray selection interface, i.e. the ``sel`` and ``isel`` methods.
    """
    # Let new containers be xarray-based
    _NEW_CONTAINER_CLS = XrDataContainer

    # Configuration options for this group ....................................
    # Whether to use deep selection by default
    LDG_ALLOW_DEEP_SELECTION = True

    # The dimensions of this group, i.e. the dimensions of the space covered by
    # the members of this group.
    LDG_DIMS = tuple()

    # How to extract coordinates of the members; for available modes, see the
    # `dantro.utils.coords.extract_coords` function.
    LDG_EXTRACT_COORDS_FROM = 'data'

    # Configuration for mode 'attrs' . . . . . . . . . . . . . . . . . . . . . 
    LDG_COORDS_ATTR_PREFIX = 'ext_coords__'
    LDG_COORDS_MODE_ATTR_PREFIX = 'ext_coords_mode__'
    LDG_COORDS_MODE_DEFAULT = 'scalar'
    LDG_STRICT_ATTR_CHECKING = False

    # Configuration for mode 'name' . . . . . . . . . . . . . . . . . . . . . .
    LDG_COORDS_SEPARATOR_IN_NAME = ';'

    # .........................................................................

    def __init__(self, *args, dims: TDims=None,
                 allow_deep_selection: bool=None, **kwargs):
        """Initialize a LabelledDataGroup
        
        Args:
            *args: Passed on to
                :py:class:`~dantro.groups.ordered.OrderedDataGroup`
            dims (TDims, optional): The dimensions associated with this group.
                If not given, will use those defined in the ``LDG_DIMS`` class
                variable. These can *not* be changed afterwards!
            allow_deep_selection (bool, optional): Whether to allow deep
                selection. If not given, will use the
                ``LDG_ALLOW_DEEP_SELECTION`` class variable's value. Behaviour
                can be changed via the property of the same name.
            **kwargs: Passed on to
                :py:class:`~dantro.groups.ordered.OrderedDataGroup`
        """
        super().__init__(*args, **kwargs)

        self._dims = dims if dims is not None else tuple(self.LDG_DIMS)
        self.__member_map = None

        self._allow_deep_selection = self.LDG_ALLOW_DEEP_SELECTION
        if allow_deep_selection is not None:
            self._allow_deep_selection = allow_deep_selection


    # Dimension and coordinates ...............................................

    @property
    def dims(self) -> Tuple[str]:
        """The names of the group-level dimensions this group manages.

        It _may_ contain dimensions that overlap with dimension names from the
        members; this is intentional.
        """
        return self._dims

    @property
    def ndim(self) -> int:
        """The rank of the space covered by the group-level dimensions."""
        return len(self.dims)

    @property
    def coords(self) -> Dict[str, List[TCoord]]:
        """Returns a dict-like container of group-level coordinates.

        The coordinates are calculated by iterating over all members and
        aggregating their individual coordinates. Once the member map is
        available, information is retrieved from there rather than
        recalculating it.
        """
        if self.member_map_available:
            return self.member_map.coords

        # Need to collect them from the members; set guarantee uniqueness
        coords = {dim_name: set() for dim_name in self.dims}

        for cont_name, cont in self.items():
            cont_coords = self._get_coords_of(cont)

            for dim_name, coord_vals in cont_coords.items():
                coords[dim_name].update(coord_vals)

        # Convert to dict of lists
        return {dim_name: sorted(list(s)) for dim_name, s in coords.items()}

    @property
    def shape(self) -> Tuple[int]:
        """Return the shape of the space covered by the group-level dimensions.

        This will be calculated from the available coordinates. Once the
        member map is available, information is retrieved from there rather
        than recalculating it.
        """
        if self.member_map_available:
            return self.member_map.shape

        # Need to derive it from the coordinates
        coords = self.coords
        return tuple([len(coords[dim_name]) for dim_name in self.dims])
    

    # Additional properties ...................................................

    @property
    def allow_deep_selection(self) -> bool:
        """Whether deep selection is allowed."""
        return self._allow_deep_selection

    @allow_deep_selection.setter
    def allow_deep_selection(self, val: bool):
        """Change whether deep selection is allowed."""
        self._allow_deep_selection = val

    @property
    def member_map(self) -> xr.DataArray:
        """Returns an array that represents the space that the members of this
        group span, where each value (i.e. a specific coordinate combination)
        is the name of the corresponding member of this group.

        Upon first call, this is computed here. If members are added, it is
        tried to accomodate them in there; if not possible, the cache will be
        invalidated.

        The member map _may_ include empty strings, i.e. coordinate
        combinations that are not covered by any member. Also, they can contain
        duplicate names, as one member can cover multiple coordinates.

        .. note::

            The member map is invalidated when new members are added that can
            not be accomodated in it. It will be recalculated when needed.
        """
        if self.member_map_available:
            return self.__member_map

        # Create an empty DataArray of strings, using the existing dimension
        # names and coordinates to label it
        mm = xr.DataArray(data=np.zeros(self.shape, dtype='<U255'),
                          dims=self.dims, coords=self.coords)

        # Iterate over members and populate the array with member names
        for name, cont in self.items():
            coords = self._get_coords_of(cont)
            # These coordinates describe a hypercube in coordinate space that
            # is to be associated with this container. Thus, the member map
            # should contain the name of the member for all these coordinates:
            mm.loc[coords] = name

        # Cache the map and return it
        self.__member_map = mm
        return mm

    @property
    def member_map_available(self) -> bool:
        """Whether the member map is available yet."""
        return (self.__member_map is not None)
    

    # Selection interface .....................................................

    def isel(self, indexers: dict=None, *, drop: bool=False,
             combination_method: str='try_concat', deep: bool=None,
             **indexers_kwargs) -> xr.DataArray:
        """Return a new labelled `xr.DataArray` with an index-selected subset
        of members of this group.
        
        If deep selection is activated, those indexers that are not available
        in the group-managed dimensions are looked up in the members of this
        group.
        
        Args:
            indexers (dict, optional): A dict with keys matching dimensions and
                values given by scalars, slices or arrays of tick indices.
                As `xr.DataArray.sel`, uses pandas-like indexing, i.e.: slices
                include the terminal value.
            drop (bool, optional): Drop coordinate variables instead of making
                them scalar.
            combination_method (str, optional): How to combine group-level data
                with member-level data. Can be:
        
                    * ``concat``: Concatenate. This can preserve the dtype, but
                        requires that no data is missing.
                    * ``merge``: Merge, using `xarray.merge`. This leads to a
                        type conversion to ``float64``, but allows members
                        being missing or coordinates not fully filling the 
                        available space.
                    * ``try_concat``: Try concatenation, fall back to merging
                        if that was unsuccessful.
        
            deep (bool, optional): Whether to allow deep indexing, i.e.: that
                ``indexers`` may contain dimensions that don't refer to group-
                level dimensions but to dimensions that are only availble among
                the member data. If ``None``, will use the value returned by
                the ``allow_deep_selection`` property.
            **indexers_kwargs: Additional indexers
        
        Returns:
            xr.DataArray: The selected data, potentially a combination of data
                on group level and member-level data.
        """
        idxrs, deep_idxrs = self._parse_indexers(indexers, allow_deep=deep,
                                                 **indexers_kwargs)

        # Use the (shallow) indexers to select (by index) those members that
        # are to be combined ...
        tbc = self.member_map.isel(idxrs, drop=drop)

        # If only a single item remains, pass deep indexers on to it
        if tbc.size == 1:
            cont = self[tbc.item()]

            if not deep_idxrs:
                return cont
            return cont.isel(deep_idxrs, drop=drop)
        
        # Now, combine them, potentially also applying deep indexing
        return self._combine(tbc, combination_method=combination_method,
                             deep_indexers=deep_idxrs, by_index=True,
                             drop=drop)

    def sel(self, indexers: dict=None, *, method: str=None,
            tolerance: float=None, drop: bool=False,
            combination_method: str='try_concat', deep: bool=None,
            **indexers_kwargs) -> xr.DataArray:
        """Return a new labelled `xr.DataArray` with a coordinate-selected
        subset of members of this group.

        If deep selection is activated, those indexers that are not available
        in the group-managed dimensions are looked up in the members of this
        group.
        
        Args:
            indexers (dict, optional): A dict with keys matching dimensions and
                values given by scalars, slices or arrays of tick labels.
                As `xr.DataArray.sel`, uses pandas-like indexing, i.e.: slices
                include the terminal value.
            method (str, optional): Method to use for inexact matches
            tolerance (float, optional): Maximum (absolute) distance between
                original and given label for inexact matches.
            drop (bool, optional): Drop coordinate variables instead of making
                them scalar.
            combination_method (str, optional): How to combine group-level data
                with member-level data. Can be:

                    * ``concat``: Concatenate. This can preserve the dtype, but
                        requires that no data is missing.
                    * ``merge``: Merge, using `xarray.merge`. This leads to a
                        type conversion to ``float64``, but allows members
                        being missing or coordinates not fully filling the 
                        available space.
                    * ``try_concat``: Try concatenation, fall back to merging
                        if that was unsuccessful.

            deep (bool, optional): Whether to allow deep indexing, i.e.: that
                ``indexers`` may contain dimensions that don't refer to group-
                level dimensions but to dimensions that are only availble among
                the member data. If ``None``, will use the value returned by
                the ``allow_deep_selection`` property.
            **indexers_kwargs: Additional indexers
        
        Returns:
            xr.DataArray: The selected data, potentially a combination of data
                on group level and member-level data.
        """
        idxrs, deep_idxrs = self._parse_indexers(indexers, allow_deep=deep,
                                                 **indexers_kwargs)

        # Use the (shallow) indexers to select (by label) those members that
        # are to be combined ...
        tbc = self.member_map.sel(idxrs, method=method,
                                  tolerance=tolerance, drop=drop)

        # If only a single item remains, pass deep indexers on to it
        if tbc.size == 1:
            cont = self[tbc.item()]

            if not deep_idxrs:
                return cont
            return cont.sel(deep_idxrs, method=method,
                            tolerance=tolerance, drop=drop)
        
        # Now, combine them, potentially also applying deep indexing
        return self._combine(tbc, combination_method=combination_method,
                             deep_indexers=deep_idxrs, by_index=False,
                             method=method, tolerance=tolerance, drop=drop)

    # Helpers .................................................................
    # General . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

    def _get_coords_of(self, obj: AbstractDataContainer, *,
                       mode=None) -> TCoordsDict:
        """Extract the coordinates for the given object using the
        `dantro.utils.coords.extract_coords` function.
        
        Args:
            obj (AbstractDataContainer): The object to get the coordinates of.
            mode (None, optional): By which coordiante extraction mode to get
                the coordinates from the object. Can be ``attrs``, ``name``,
                ``data`` or anything else specified in
                ~`dantro.utils.coords.extract_coords`.
        
        Returns:
            TCoordsDict: The extracted coordinates
        """
        # Depending on the mode, compile the dict of additional parameters
        mode = mode if mode is not None else self.LDG_EXTRACT_COORDS_FROM
        kwargs = dict()

        if mode == 'attrs':
            kwargs['coords_attr_prefix'] = self.LDG_COORDS_ATTR_PREFIX
            kwargs['mode_attr_prefix'] = self.LDG_COORDS_MODE_ATTR_PREFIX
            kwargs['default_mode'] = self.LDG_COORDS_MODE_DEFAULT
            kwargs['strict'] = self.LDG_STRICT_ATTR_CHECKING

        elif mode == 'name':
            kwargs['separator'] = self.LDG_COORDS_SEPARATOR_IN_NAME

        return extract_coords(obj, dims=self.dims, mode=mode, **kwargs)

    def _add_container_callback(self, cont: AbstractDataContainer) -> None:
        """Called by the base class after adding a container, this method
        checks whether the member map needs to be invalidated or whether the
        new container can be accomodated in it.

        If it can be accomodated, the member map will be adjusted such that for
        all coordinates associated with the given ``cont``, the member map
        points to the newly added container.
        
        Args:
            cont (AbstractDataContainer): The newly added container
        
        Returns:
            None: Description
        """
        # First, let the parent class do its thinkg
        super()._add_container_callback(cont)

        # Don't have to do anything if there is no member map yet
        if not self.member_map_available:
            return

        # There is a map. Check if it can accomodate the new container
        coords = self._get_coords_of(cont)
        
        try:
            self.__member_map.loc[coords] = cont.name

        except:
            # Cannot accomodate it -> invalidate it
            self.__member_map = None

    # For selection . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

    def _parse_indexers(self, indexers: dict, *, allow_deep: bool,
                        **indexers_kwargs) -> Tuple[dict, dict]:
        """Parses the given indexer arguments and split them into indexers for
        the selection of group members and deep selection.
        
        Args:
            indexers (dict): The indexers dict, may be empty
            allow_deep (bool): Whether to allow deep selection
            **indexers_kwargs: Additional indexers
        
        Returns:
            Tuple[dict, dict]: (shallow indexers, deep indexers)
        
        Raises:
            ValueError: If deep indexers were given but deep selection was not
                enabled
        """
        allow_deep = (allow_deep if allow_deep is not None
                      else self.allow_deep_selection)
        idxrs = dict(**(indexers if indexers else {}), **indexers_kwargs)

        # Split by those for deep selection and those for this group
        deep_idxrs = {k: v for k, v in idxrs.items() if k not in self.dims}
        idxrs = {k: v for k, v in idxrs.items() if k in self.dims}

        if deep_idxrs and not allow_deep:
            raise ValueError("Deep indexing is not allowed for {}, but got "
                             "indexers that don't match any of its dimension "
                             "names: {}.  You can change this behavior using "
                             "the allow_deep_selection property or the class "
                             "variable LDG_ALLOW_DEEP_SELECTION."
                             "".format(self.logstr, ", ".join(self.dims)))

        return idxrs, deep_idxrs

    def _combine(self, cont_names: xr.DataArray, *, combination_method: str,
                 deep_indexers: dict, by_index: bool,
                 **sel_kwargs) -> xr.Dataset:
        """Combine the given objects by the specified method. If deep indexers
        are given, apply the deep indexing on each of the members.
        
        This method receives a labelled array of container names, on which the
        selection already took place. The aim is now to align the objects these
        names refer to, including their coordinates, and thereby construct an
        array that contains both the dimensions given by the ``cont_names``
        array and each members' data dimensions.
        
        Available combination methods are based either on `xarray.merge`
        operations or `xarray.concat` along each dimension. For both these
        combination methods, the members of this group need to be prepared such
        that the operation can be applied, i.e.: they need to already be in an
        array capable of that operation and they need to directly or indirectly
        preserve coordinate information.
        
        For that purpose, an object-array is constructed that has the same
        shape as the given ``cont_names``. As the `xarray.Dataset` and
        `xarray.DataArray` types have issues with handling array-like objects
        in object arrays, this is done via a `numpy.ndarray`.
        
        Args:
            cont_names (xr.DataArray): The pre-selected member map object, i.e.
                a labelled array containing names of the desired members that
                are to be combined.
            combination_method (str): How to combine them: concat, try_concat,
                or merge. Concatenation will allow preserving the dtype of the
                underlying data.
            deep_indexers (dict): Whether any further indexing is to take place
                before combination.
            by_index (bool): Whether the deep indexing should take place by
                index; if False, will use label-based selection.
            **sel_kwargs: Passed on to ``.sel`` or ``.isel``.
        
        Returns:
            xr.Dataset: The data of the members from ``cont_names``, combined
                using the given combination method.
        
        Raises:
            ValueError: Invalid combination method
            KeyError: In ``concat`` mode, upon missing members.
        """
        def get_cont(name: str, combination_method: str
                     ) -> Tuple[Union[XrDataContainer, None], str]:
            """Retrieve the container from the group, potentially changing the
            combination method. If no container could be found, returns None,
            which denotes that further processing should be skipped
            """
            try:
                cont = self[name]

            except KeyError as err:
                if combination_method != 'concat':
                    # Failing is ok. But cannot do anything else here
                    return None, combination_method

                # Otherwise, should raise!
                raise KeyError("Could not find a member named '{}' in {}, but "
                               "need it for concatenation! Make sure that the "
                               "member can be found under this name or change "
                               "the combination method to 'merge' or "
                               "'try_concat'.".format(name, self.logstr)
                               ) from err

            else:
                return cont, combination_method

        def process_cont(cont, coords) -> Tuple[xr.DataArray, dict]:
            """Process the given container and coordinates into a data array;
            this also applies the deep selection.
            """
            # Apply the coordinates of the overlapping dimensions
            # (Does nothing if there are no overlapping dimensions)
            darr = cont.sel({dim: coord for dim, coord in coords.items()
                             if dim in cont.dims},
                             drop=False)
            # This is to ensure that the array that is used matches only a
            # single coordinate combination, i.e. one _point_ in the space
            # spanned by self.member_map.coords.
            # This operation _might_ increase memory usage temporarily, because
            # data from the same member is accessed multiple times and only one
            # coordinate combination is extracted from it (instead of selecting
            # all at once; which would however require a different architecture
            # and not allow using the convenient xarray interface).
            
            # Apply the deep indexers
            if by_index:
                darr = darr.isel(deep_indexers, **sel_kwargs)
            else:
                darr = darr.sel(deep_indexers, **sel_kwargs)

            # For the following, the container coordinates may not contain
            # any dimension names that are overlapping with those of the group
            coords = {dim_name: coords for dim_name, coords in coords.items()
                      if dim_name not in darr.dims}

            return darr, coords


        dsets = np.zeros(cont_names.shape, dtype='object')
        dsets.fill(dict())  # placeholders, ignored in xr.merge

        # Create an iterator over the container names (mirrors dsets iteration)
        names_iter = np.nditer(cont_names, flags=('multi_index', 'refs_ok'))

        for name in names_iter:
            # Get the corresponding member container, potentially changing the
            # combination method
            cont, combination_method = get_cont(name.item(),
                                                combination_method)

            # Might not have been found; go to the next iteration
            if cont is None:
                continue

            # Get the coordinates for this member container and further process
            # the container into a DataArray
            coords = cont_names[names_iter.multi_index].coords
            darr, coords = process_cont(cont, coords)

            # As it's easier to work on xr.Datasets than on xr.DataArray-like
            # objects, create a dataset from the container, using a temporary
            # name which will later be used to resolve it back to a DataArray
            dset = darr.to_dataset(name='_tmp_dset_name')

            # Now, need to expand the dimensions to accomodate the coordinates.
            # Add the new dimensions in front. (Important for concatenation!)
            dset = dset.expand_dims(dim=list(coords.keys()))
            # NOTE While this creates a non-shallow copy of the data, there is
            #      no other way of doing this: a copy can only be avoided if
            #      the DataArray can re-use the existing variables – for the
            #      changes it needs to do to expand the dims, however, it will
            #      necessarily need to create a copy of the original data.
            #      Thus, we might as well let xarray take care of that instead
            #      of bothering with that ourselves ...

            # ...and assign coordinates to them (shallow copy of existing dset)
            dset = dset.assign_coords(**{k: [v] for k, v in coords.items()})

            # Done. Store it in the object-array of datasets
            dsets[names_iter.multi_index] = dset

        # Now ready to combine them.
        if combination_method == 'concat':
            dset = self._combine_by_concatenation(dsets,
                                                  dims=cont_names.dims)

        elif combination_method == 'merge':
            dset = self._combine_by_merge(dsets)

        elif combination_method == 'try_concat':
            try:
                dset = self._combine_by_concatenation(dsets,
                                                      dims=cont_names.dims)

            except Exception as exc:
                # NOTE The exception is now something other than a member
                #      missing, i.e. some numerical issue during concatenation
                # Try again with merging ...
                log.warning("Failed concatenation with %s: %s",
                            exc.__class__.__name__, exc)
                dset = self._combine_by_merge(dsets)

        else:
            raise ValueError("Invalid combination_method argument: {}! "
                             "Available methods: try_concat, concat, merge."
                             "".format(combination_method))

        # Combined into one dataset now, with '_tmp_dset_name' data variable...
        # Convert back into a DataArray; can drop the temporary name now.
        darr = dset['_tmp_dset_name']
        darr.name = None
        return darr

    @classmethod
    def _combine_by_merge(cls, dsets: np.ndarray) -> xr.Dataset:
        """Combine the given datasets by merging using `xarray.merge`.
        
        Args:
            dsets (np.ndarray): The object-dtype array of xr.Datasets that are
                to be combined.
        
        Returns:
            xr.Dataset: All datasets, aligned and combined via `xarray.merge`
        """
        log.debug("Combining %d datasets by merging ...", dsets.size)

        dset = xr.merge(dsets.flat)

        log.debug("Merge successful.")
        return dset
    
    @classmethod
    def _combine_by_concatenation(cls, dsets: np.ndarray, *,
                                  dims: TDims) -> xr.Dataset:
        """Combine the given datasets by concatenation using `xarray.concat`
        and subsequent application along all dimensions specified in ``dims``.
        
        Args:
            dsets (np.ndarray): The object-dtype array of xr.Dataset objects
                that are to be combined by concatenation.
            dims (TDims): The dimension names corresponding to _all_ the
                dimensions of the ``dsets`` array.
        
        Returns:
            xr.Dataset: The dataset resulting from the concatenation
        """
        log.debug("Combining %d datasets by concatenation along %d "
                  "dimension%s ...", dsets.size, len(dsets.shape),
                  "s" if len(dsets.shape) != 1 else "")

        # Go over all dimensions and concatenate
        # This effectively reduces the dsets array by one dimension in each
        # iteration by applying the xr.concat function along the axis
        # NOTE np.apply_along_axis would be what is desired here, but that
        #      function unfortunately tries to cast objects to np.arrays
        #      which is not what we want here at all!
        #      Thus, there is one implemented in dantro.tools ...
        for dim_idx, dim_name in reversed(list(enumerate(dims))):
            log.debug("Concatenating along axis '%s' (axis # %d) ...",
                      dim_name, dim_idx)

            dsets = apply_along_axis(xr.concat, axis=dim_idx, arr=dsets,
                                     dim=dim_name)

        log.debug("Concatenation successful.")

        # The single item in the now scalar array is the combined xr.Dataset
        return dsets.item()



"""Defines a loader mixin to load plain text objects"""

from ..containers import StringContainer
from ._tools import add_loader

# -----------------------------------------------------------------------------

class TextLoaderMixin:
    """Supplies functionality to load plain text files"""

    @add_loader(TargetCls=StringContainer)
    def _load_plain_text(filepath: str, *, TargetCls: type,
                         **load_kwargs) -> StringContainer:
        """Loads the content of a plain text file back into a
        :py:class:`~dantro.containers.general.StringContainer`.
        
        Args:
            filepath (str): Where the plain text file is located
            TargetCls (type): The class constructor
            **load_kwargs: Passed on to ``open``, see there for possible kwargs
        
        Returns:
            StringContainer: The reconstructed StringContainer
        """
        with open(filepath, **load_kwargs) as f:
            data = f.read()

        return TargetCls(data=data, attrs=dict(filepath=filepath))

    # Also make it available under plain 'text'
    _load_text = _load_plain_text



"""In this module, the BaseDataGroup is specialized for holding members in a
specific order.
"""

import logging
import collections
from typing import Generator, List

from ..base import BaseDataGroup
from ..mixins import IntegerItemAccessMixin
from ..utils import IntOrderedDict

# Local constants
log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------

class OrderedDataGroup(BaseDataGroup, collections.abc.MutableMapping):
    """The OrderedDataGroup class manages groups of data containers, preserving
    the order in which they were added to this group.

    It uses an OrderedDict to associate containers with this group.
    """
    # Use OrderedDict for storage in insertion order
    _STORAGE_CLS = collections.OrderedDict

# -----------------------------------------------------------------------------

class IndexedDataGroup(IntegerItemAccessMixin, OrderedDataGroup):
    """The IndexedDataGroup class holds members that are of the same type and
    have names that can directly be interpreted as positive integers.

    Especially, this group maintains the correct order of members according to
    integer ordering.

    .. note::

      Albeit the members of this group are ordered, item access still refers
      to the _names_ of the members, not their index within the sequence!
    """
    # Use an orderable dict for storage, i.e. something like OrderedDict, but
    # where it's not sorted by insertion order but by key.
    _STORAGE_CLS = IntOrderedDict

    # The child class should not necessarily be of the same type as this class.
    _NEW_GROUP_CLS = OrderedDataGroup


    # Advanced key access .....................................................
    
    def key_at_idx(self, idx: int) -> str:
        """Get a key by its index within the container. Can be negative.
        
        Args:
            idx (int): The index within the member sequence
        
        Returns:
            str: The desired key
        
        Raises:
            IndexError: Index out of range
        """
        # Imitate indexing behaviour of lists, tuples, ...
        if not isinstance(idx, int):
            raise TypeError("Expected integer, got {} '{}'!"
                            "".format(type(idx), idx))

        if idx >= len(self) or idx < -len(self):
            raise IndexError("Index {:d} out of range for {} with {} members!"
                             "".format(idx, self.logstr, len(self)))

        # Wraparound negative
        idx = idx if idx >= 0 else idx%len(self)

        for i, k in enumerate(self.keys()):
            if i == idx:
                return k

    def keys_as_int(self) -> Generator[int, None, None]:
        """Returns an iterator over keys as integer values"""
        for k in self.keys():
            yield int(k)


    # Customizations of parent methods ........................................

    def _ipython_key_completions_(self) -> List[int]:
        """For ipython integration, return a list of available keys.

        Unlike the BaseDataGroup method, which returns a list of strings, this
        returns a list of integers.
        """
        return list(self.keys_as_int())



"""This sub-package implements non-abstract dantro plot creator classes"""

# Supply the base class
from dantro.plot_creators.pcr_base import BasePlotCreator

# And all derived PlotCreator classes
from dantro.plot_creators.pcr_ext import ExternalPlotCreator
from dantro.plot_creators.pcr_decl import DeclarativePlotCreator
from dantro.plot_creators.pcr_psp import UniversePlotCreator
from dantro.plot_creators.pcr_psp import MultiversePlotCreator
from dantro.plot_creators.pcr_vega import VegaPlotCreator

# Make some associated objects easier to import
from dantro.plot_creators.pcr_ext import is_plot_func
from dantro.plot_creators.pcr_ext import PlotHelper

# And gather them into a dictionary that gives names to each of them
ALL = dict(external=ExternalPlotCreator,
           declarative=DeclarativePlotCreator,
           universe=UniversePlotCreator,
           multiverse=MultiversePlotCreator,
           vega=VegaPlotCreator,
           )



"""The groups sub-package implements BaseDataGroup specializations"""

from .ordered import OrderedDataGroup, IndexedDataGroup
from .labelled import LabelledDataGroup
from .time_series import TimeSeriesGroup, HeterogeneousTimeSeriesGroup
from .pspgrp import ParamSpaceStateGroup, ParamSpaceGroup
from .network import NetworkGroup



"""Implements the Link class"""

import weakref
import logging
from typing import TypeVar

from ..abc import PATH_JOIN_CHAR
from ..base import BaseDataContainer, BaseDataGroup
from ..mixins import ForwardAttrsMixin

# Local constants
log = logging.getLogger(__name__)

# Type definitions
TGroupOrContainer = TypeVar('TGroupOrContainer',
                            BaseDataContainer, BaseDataGroup)

# -----------------------------------------------------------------------------

class Link(ForwardAttrsMixin):
    """A link is a connection between two objects in the data tree, i.e. a
    data group and a data container.

    It has a source object that it is coupled to and a relative path from that
    object to the target object.

    Whenever attribute access occurs, an object of this class will resolve the
    linked object (if not already cached) and then forward the attribute call
    to that object.
    """
    FORWARD_ATTR_TO = "target_object"  # ...but see …_forwarding_target method!

    def __init__(self, *, anchor: TGroupOrContainer, rel_path: str):
        """Initialize a link from an anchor and a relative path to a target"""
        # Use name-mangling to not take up any attributes that might get in the
        # way of attribute forwarding ...
        self.__anchor = weakref.ref(anchor)
        self.__rel_path = rel_path
        self.__target_ref_cache = None

        log.debug("Created link with anchor %s and relative path '%s'.",
                  anchor.logstr, rel_path)

    @property
    def target_weakref(self) -> weakref:
        """Resolve the target and return the weak reference to it"""
        if self.__target_ref_cache is None:
            self.__resolve_target_ref()
        return self.__target_ref_cache

    @property
    def target_object(self) -> TGroupOrContainer:
        """Return a (non-weak) reference to the actual target object"""
        return self.target_weakref()  # calling property to resolve the weakref

    @property
    def anchor_weakref(self) -> weakref:
        """Resolve the weak reference to the anchor and return it, i.e.:
        return a reference to the actual object.
        """
        return self.__anchor

    @property
    def anchor_object(self) -> TGroupOrContainer:
        """Return a (non-weak) reference to the anchor object"""
        return self.__anchor()

    @property
    def target_rel_path(self) -> str:
        """Returns the relative path to the target"""
        return self.__rel_path

    def __resolve_target_ref(self) -> None:
        """Resolves the weak reference to the target object and caches it"""
        # Start at the anchor, resolving the weak reference via '()'
        obj = self.__anchor()

        # If there is no relative path given, this is simply a link to itself
        if not self.__rel_path:
            self.__target_ref_cache = self.__anchor
            return obj

        # Distinguish between anchors that are groups and anchors that are
        # containers; the latter need to be embedded and the _parent_ object is
        # then where the path should start from.
        if not isinstance(obj, BaseDataGroup):
            # Assume it's a container-like anchor, although it might also be a
            # whole other type. We don't care as long as a parent is defined.
            if obj.parent is None:
                raise ValueError("The anchor object {} is not embedded into a "
                                 "data tree; cannot resolve the target '{}'! "
                                 "Either choose a group-like dantro object as "
                                 "an anchor or embed the container-like "
                                 "object into a group."
                                 "".format(obj.logstr,
                                           self.__rel_path))

            obj = obj.parent

        # Try traversing the path, going to the parent or a child depending on
        # which kind of segment is encountered.
        try:
            for segment in self.__rel_path.split(PATH_JOIN_CHAR):
                # Skip empty segments, i.e. `foo//bar` paths. This makes path
                # traversal more robust and mirrors UNIX behaviour (try it).
                if not segment:
                    continue

                obj = obj.parent if segment == ".." else obj[segment]

        except Exception as err:
            raise ValueError("Failed resolving target of link '{}' relative "
                             "to anchor {} @ {}. Are anchor and target part "
                             "of the same data tree?"
                             "".format(self.__rel_path,
                                       self.__anchor().logstr,
                                       self.__anchor().path)
                             ) from err

        log.debug("Resolved link '%s' relative to anchor %s @ %s",
                  self.__rel_path, self.__anchor().logstr,
                  self.__anchor().path)

        self.__target_ref_cache = weakref.ref(obj)

    def _forward_attr_get_forwarding_target(self):
        """Get the object that the attribute call is to be forwarded to, i.e.
        the resolved target object. This invokes resolution of the target and
        caching of the corresponding weakref, but the returned (strong) ref
        will not be cached.
        """
        return self.target_object



#!/usr/bin/env python3

from setuptools import setup, find_packages

# Dependencies for dantro itself
install_deps = [
    'numpy>=1.14',
    'xarray>=0.15.0',
    'dask>=2.10.1',
    'toolz>=0.10.0',        # Needed for dask.delayed
    'distributed>=2.10.0',  # Needed for dask's distributed scheduler
    'scipy>=1.4.1',         # Used as a netcdf4 storage engine for xarray
    'h5py>=2.7.0',
    'networkx>=2.2',
    'ruamel.yaml>=0.16.5',
    'matplotlib>=3.1.3',
    'paramspace>=2.2.3'
    ]

# Dependencies for the tests
test_deps = ['pytest>=3.4.0', 'pytest-cov>=2.5.1', 'tox>=3.1.2']

# Dependencies for the documentation
doc_deps = ['sphinx>=2.4,<3.0', 'sphinx_rtd_theme']

# .............................................................................

DESCRIPTION = "Loads, handles, and plots hierarchically structured data"
LONG_DESCRIPTION = """
With dantro, hierarchically structured heterogeneous data can be loaded into a
uniform data structure, a so-called data tree. Furthermore, dantro provides
capabilities to generically work on and create plots from that data. The whole
package aims at implementing an abstract and homogeneous interface that allows
specializing the provided classes to the needs to the data that is to be worked
with. At the same time, it already provides useful functionality, which makes
specialization easier.

Learn More
----------

For more information, visit the project page and have a look at the README:
https://ts-gitlab.iup.uni-heidelberg.de/utopia/dantro

"""


# .............................................................................

# A function to extract version number from __init__.py
def find_version(*file_paths) -> str:
    """Tries to extract a version from the given path sequence"""
    import os, re, codecs

    def read(*parts):
        """Reads a file from the given path sequence, relative to this file"""
        here = os.path.abspath(os.path.dirname(__file__))
        with codecs.open(os.path.join(here, *parts), 'r') as fp:
            return fp.read()

    # Read the file and match the __version__ string
    file = read(*file_paths)
    match = re.search(r"^__version__\s?=\s?['\"]([^'\"]*)['\"]", file, re.M)
    if match:
        return match.group(1)
    raise RuntimeError("Unable to find version string in " + str(file_paths))


# .............................................................................

setup(
    name='dantro',
    #
    # Set the version from dantro.__version__
    version=find_version('dantro', '__init__.py'),
    #
    # Project info
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    #
    author="dantro developers",
    author_email="dantro-dev@iup.uni.heidelberg.de",
    licence='LGPL-3.0-or-later',
    url='https://ts-gitlab.iup.uni-heidelberg.de/utopia/dantro',
    classifiers=[
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Utilities',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)'
    ],
    #
    # Distribution details, dependencies, ...
    packages=find_packages(exclude=["tests.*", "tests"]),
    install_requires=install_deps,
    tests_require=test_deps,
    test_suite='py.test',
    extras_require=dict(test_deps=test_deps, doc_deps=doc_deps)
)



"""This implements the ParamSpaceGroup plot creators, based on the
ExternalPlotCreator and providing additional functionality for data that is
stored in a ParamSpaceGroup.
"""

import copy
import logging
from typing import Union, Tuple, Callable, Sequence

import numpy as np
import xarray as xr

from paramspace import ParamSpace

from .pcr_ext import ExternalPlotCreator
from ..groups import ParamSpaceGroup, ParamSpaceStateGroup
from ..tools import recursive_update
from ..abc import PATH_JOIN_CHAR
from ..dag import TransformationDAG, DAGReference, DAGTag, DAGNode

# Local constants
log = logging.getLogger(__name__)


# -----------------------------------------------------------------------------

class MultiversePlotCreator(ExternalPlotCreator):
    """A MultiversePlotCreator is an ExternalPlotCreator that allows data to be
    selected before being passed to the plot function.
    """
    # Where the ``ParamSpaceGroup`` object is expected within the data manager
    PSGRP_PATH = None

    # .........................................................................

    def __init__(self, *args, psgrp_path: str=None, **kwargs):
        """Initialize a MultiversePlotCreator"""
        super().__init__(*args, **kwargs)

        if psgrp_path:
            self.PSGRP_PATH = psgrp_path

    @property
    def psgrp(self) -> ParamSpaceGroup:
        """Retrieves the parameter space group associated with this plot
        creator by looking up a certain path in the data manager.
        """
        if self.PSGRP_PATH is None:
            raise ValueError("Missing class variable PSGRP_PATH! Either set "
                             "it directly or pass the `psgrp_path` argument "
                             "to the __init__ function.")

        # Retrieve the parameter space group
        return self.dm[self.PSGRP_PATH]

    def _prepare_plot_func_args(self, *args,
                                select: dict=None,
                                select_and_combine: dict=None,
                                **kwargs) -> Tuple[tuple, dict]:
        """Prepares the arguments for the plot function.
        
        This also implements the functionality to select and combine data from
        the Multiverse and provide it to the plot function. It can do so via
        the associated :py:class:`~dantro.groups.pspgrp.ParamSpaceGroup`
        directly or by creating a :py:class:`~dantro.dag.TransformationDAG`
        that leads to the same results.
        
        .. warning::
        
            The ``select_and_combine`` argument behaves slightly different to
            the ``select`` argument! In the long term, the ``select`` argument
            will be deprecated.
        
        Args:
            *args: Positional arguments to the plot function.
            select (dict, optional): If given, selects and combines multiverse
                data using
                :py:meth:`~dantro.groups.pspgrp.ParamSpaceGroup.select`.
                The result is an ``xr.Dataset`` and it is made available to
                the plot function as ``mv_data`` argument.
            select_and_combine (dict, optional): If given, interfaces with the
                DAG to select, transform, and combine data from the multiverse
                via the DAG.
            **kwargs: Keyword arguments for the plot function. If DAG usage is
                enabled, these contain further arguments like ``transform``
                that are filtered out accordingly.
        
        Returns:
            Tuple[tuple, dict]: The (args, kwargs) tuple for calling the plot
                function. These now include either the DAG results or the
                additional ``mv_data`` key.
        
        Raises:
            TypeError: If both or neither of the arguments ``select`` and/or
                ``select_and_combine`` were given.
        """
        # Distinguish between the new DAG-based selection interface and the
        # old (and soon-to-be-deprecated) one.
        if select and not select_and_combine:
            # Select multiverse data via the ParamSpaceGroup
            kwargs['mv_data'] = self.psgrp.select(**select)

        elif select_and_combine:
            # Pass both arguments along
            kwargs['select'] = select
            kwargs['select_and_combine'] = select_and_combine

        else:
            raise TypeError("Expected at least one of the arguments `select` "
                            "or `select_and_combine`, got neither!")

        # Let the parent method (from ExternalPlotCreator) do its thing.
        # It will invoke the specialized _get_dag_params and _create_dag helper
        # methods that are implemented by this class.
        return super()._prepare_plot_func_args(*args, **kwargs)

    # .........................................................................
    # DAG specialization

    def _get_dag_params(self, *, select_and_combine: dict, **cfg
                        ) -> Tuple[dict, dict]:
        """Extends the parent method by extracting the select_and_combine
        argument that handles MultiversePlotCreator behaviour
        """
        dag_params, plot_kwargs = super()._get_dag_params(**cfg)

        # Add the select_and_combine argument; converting None to an empty dict
        select_and_combine = select_and_combine if select_and_combine else {}
        dag_params['init']['select_and_combine'] = select_and_combine

        return dag_params, plot_kwargs

    def _create_dag(self, *, _plot_func: Callable,
                    select_and_combine: dict,
                    select: dict=None, transform: Sequence[dict]=None,
                    select_base: str=None, select_path_prefix: str=None,
                    **dag_init_params) -> TransformationDAG:
        """Extends the parent method by translating the ``select_and_combine``
        argument into selection of tags from a universe subspace, subsequent
        transformations, and a ``combine`` operation, that aligns the data in
        the desired fashion.
        
        This way, the :py:meth:`~dantro.groups.pspgrp.ParamSpaceGroup.select`
        method's behaviour is emulated in the DAG.
        
        Args:
            _plot_func (Callable): The plot function; passed on to parent
                method, where tag availability is checked.
            select_and_combine (dict): The parameters to define which data from
                the universes to select and combine before applying further
                transformations.
            select (dict, optional): Additional select operations; these are
                *not* applied to *each* universe but only globally, after the
                ``select_and_combine`` nodes are added.
            transform (Sequence[dict], optional): Additional transform
                operations that are added to the DAG after both the
                ``select_and_combine``- and ``select``-related transformations
                were added.
            select_base (str, optional): The select base for the ``select``
                argument. These are *not* relevant for the selection that
                occurs via the ``select_and_combine`` argument and is only set
                after all ``select_and_combine``-related transformations are
                added to the DAG.
            select_path_prefix (str, optional): The selection path prefix for
                the ``select`` argument. Cannot be used here.
            **dag_init_params: Further initialization arguments to the DAG.
        
        Returns:
            TransformationDAG: The populated DAG object.
        """
        def add_uni_transformations(dag: TransformationDAG, *,
                                    uni: ParamSpaceStateGroup,
                                    coords: dict, path: str,
                                    transform: Sequence[dict]=None,
                                    **select_kwargs) -> DAGReference:
            """Adds the sequence of select and transform operations that is
            to be applied to the data from a *single* universe; this is in
            preparation to the combination of all single-universe DAG strands.
            The last transformation node that is added by this helper is the
            one that is used as input to the combination methods.
            
            The easiest way to add a sequence of transformations that is based
            on a selection from the DataManager is to use TransformationDAG's
            :py:meth:`~dantro.dag.TransformationDAG.add_nodes` method. To that
            end, this helper function creates the necessary parameters for the
            ``select`` argument to that method.

            .. note::

                To not crowd the tag space, tags are omitted on these transform
                operations, unless manually specified.
            """
            # Create the full path that is needed to get from the selection
            # base (the ParamSpaceGroup) to the desired path within the
            # current universe.
            field_path = PATH_JOIN_CHAR.join([uni.name, path])
            
            # Prepare arguments for the select operation
            select = dict()
            select[uni.name] = dict(path=field_path, transform=transform,
                                    omit_tag=True, **select_kwargs)
            
            # Add the nodes that handle the selection and optional transform
            # operations on the selected data. This is all user-determined.
            # The selection base is the DataManager.
            dag.add_nodes(select=select)

            # With the latest-added transformation as input, add the parameter
            # space coordinates to it such that all single universes can be
            # aligned properly. Don't cache this result.
            return dag.add_node(operation='dantro.expand_dims',
                                args=[DAGNode(-1)],
                                kwargs=dict(
                                    dim={k:[v] for k,v in coords.items()}),
                                file_cache=dict(read=False, write=False))

        def add_transformations(dag: TransformationDAG, *,
                                path: str, tag: str, subspace: dict,
                                combination_method: str,
                                combination_kwargs: dict=None,
                                **select_kwargs) -> None:
            """Adds the sequence of transformations that is necessary to select
            data from a single universe and transform it, i.e.: all the preps
            necessary to arrive at another input argument to the combiation.
            """
            def get_uni(state_no: int) -> ParamSpaceStateGroup:
                """Given a state number, returns the corresponding universe"""
                try:
                    return self.psgrp[state_no]
                
                except Exception as exc:
                    if combination_method not in ['merge']:
                        raise ValueError("Missing data for universe {}, which "
                                         "is required for concatenation."
                                         "".format(state_no))
                    log.warning("Missing data for universe %d; this will lead "
                                "to NaNs in the combined results.")
                    raise NotImplementedError("Cannot handle missing data "
                                              "during merge operations yet.")
                    return None  # TODO

            # Get the parameter space object
            psp = copy.deepcopy(self.psgrp.pspace)

            # Apply the subspace mask, if given
            if subspace:
                psp.activate_subspace(**subspace)

            # Prepare iterators and extract shape information
            psp_it = psp.iterator(with_info=('state_no', 'current_coords'),
                                  omit_pt=True)

            # For each universe in the subspace, add a sequence of transform
            # operations that lead to the data being selected and (optionally)
            # further transformed. Keep track of the reference to the last
            # node of each branch, which is a node that assigns coordinates to
            # each node of the parameter space
            refs = [add_uni_transformations(dag, uni=get_uni(state_no),
                                            coords=coords, path=path,
                                            **select_kwargs)
                    for state_no, coords in psp_it]

            # Depending on the chosen combination method, create corresponding
            # additional transformations for combination via merge or via
            # concatenation.
            if combination_method in ['merge']:
                dag.add_node(operation='dantro.merge',
                             args=[refs],
                             kwargs=dict(reduce_to_array=True),
                             tag=tag,
                             **(combination_kwargs if combination_kwargs
                                else {}))

            elif combination_method in ['concat']:
                # For concatenation, it's best to have the data in an ndarray
                # of xr.DataArray's, such that sequential applications of the
                # xr.concat method along the array axes can be used for
                # combining the data.
                dag.add_node(operation='populate_ndarray',
                             args=[*refs],
                             kwargs=dict(shape=psp.shape, dtype='object'))

                dag.add_node(operation='dantro.multi_concat',
                             args=[DAGNode(-1)],
                             kwargs=dict(dims=list(psp.dims.keys())),
                             tag=tag,
                             **(combination_kwargs if combination_kwargs
                                else {}))

            else:
                raise ValueError("Invalid combination method '{}'! Available "
                                 "methods: merge, concat."
                                 "".format(combination_method))

        def add_sac_transformations(dag: TransformationDAG, *,
                                    fields: dict, subspace: dict=None,
                                    combination_method: str='concat',
                                    base_path: str=None) -> None:
            """Adds transformations to the given DAG that select data from the
            selected multiverse subspace.
            
            Args:
                dag (TransformationDAG): The DAG to add nodes to that represent
                    the select-and-combine operations.
                fields (dict): Which fields to select from the separate
                    universes.
                subspace (dict, optional): The (default) subspace to select the
                    data from.
                combination_method (str, optional): The (default) combination
                    method of the multidimensional data.
                base_path (str, optional): If given, ``path`` specifications
                    of each field can be seen as relative to this path.
            """
            # To make selections shorter, add a transformation to get to the
            # ParamSpaceGroup and set that as the selection base.
            dag.select_base = dag.add_node(operation='getitem',
                                           args=[DAGTag('dm'),self.PSGRP_PATH])

            # For all tags, update the default values with custom arguments
            # and then add transformations
            for tag, spec in fields.items():
                # For safety, work on a copy
                spec = copy.deepcopy(spec)

                # If a base path was given, prepend it. This is the path that
                # is selected *within* each universe.
                if base_path is not None:
                    spec['path'] = PATH_JOIN_CHAR.join([base_path,
                                                        spec['path']])

                # Parse parameters, i.e.: Use defaults defined on this level
                # if the spec does not provide more specific information.
                spec['subspace'] = spec.get('subspace',copy.deepcopy(subspace))
                spec['combination_method'] = spec.get('combination_method',
                                                      combination_method)

                # Add the transformations for this specific tag
                add_transformations(dag, tag=tag, **spec)
        
            # Done. :)
            log.remark("Added select-and-combine transformations for "
                       "tags: %s. The DAG contains %d nodes now.",
                       ", ".join(fields.keys()), len(dag.nodes))
            # NOTE Resetting the selection base is not necessary here, because
            #      the user-specified value is set directly after this function
            #      returns.

        # . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
        # To not create further confusion regarding base paths, raise an error
        # if it is attempted to set the `select_path_prefix` argument.
        if select_path_prefix:
            raise ValueError("The select_path_prefix argument cannot be used "
                             "within {}! Use the select_and_combine.base_path "
                             "argument instead.".format(self.logstr))

        # Initialize an (empty) DAG, i.e.: without select and transform args
        # and without setting the selection base
        dag = super()._create_dag(_plot_func=_plot_func, **dag_init_params)

        # Add nodes that perform the "select and combine" operations, based on
        # selections from the DataManager
        add_sac_transformations(dag, **select_and_combine)

        # Can now set the selection base to the user-intended value and add
        # the other user-specified transformations
        dag.select_base = select_base
        dag.add_nodes(select=select, transform=transform)
        
        return dag


# -----------------------------------------------------------------------------

class UniversePlotCreator(ExternalPlotCreator):
    """A UniversePlotCreator is an ExternalPlotCreator that allows looping of
    all or a selected subspace of universes.
    """
    # Where the `ParamSpaceGroup` object is expected within the data manager
    PSGRP_PATH = None

    # .........................................................................

    def __init__(self, *args, psgrp_path: str=None, **kwargs):
        """Initialize a UniversePlotCreator"""
        super().__init__(*args, **kwargs)

        if psgrp_path:
            self.PSGRP_PATH = psgrp_path

        # Add custom attributes
        self._without_pspace = False

        # Cache attributes
        self._psp = None
        self._psp_active_smap_cache = None

    @property
    def psgrp(self) -> ParamSpaceGroup:
        """Retrieves the parameter space group associated with this plot
        creator by looking up a certain path in the data manager.
        """
        if self.PSGRP_PATH is None:
            raise ValueError("Missing class variable PSGRP_PATH! Either set "
                             "it directly or pass the `psgrp_path` argument "
                             "to the __init__ function.")

        # Retrieve the parameter space group
        return self.dm[self.PSGRP_PATH]

    def prepare_cfg(self, *, plot_cfg: dict, pspace: Union[dict, ParamSpace]
                    ) -> Tuple[dict, ParamSpace]:
        """Converts a regular plot configuration to one that can be configured
        to iterate over multiple universes via a parameter space.

        This is implemented in the following way:
            1. Extracts the `universes` key from the configuration and parses
               it, ensuring it is a valid dict for subspace specification
            2. Creates a new ParamSpace object that additionally contains the
               parameter dimensions corresponding to the universes. These are
               stored in a _coords dict inside the returned plot configuration.
            3. Apply the parsed `universes` key to activate a subspace of the
               newly created parameter space.
            4. As a mapping from coordinates to state numbers is needed, the
               corresponding active state mapping is saved as an attribute to
               the plot creator, such that it is available later when
               the state number needs to be retrieved only be the info of the
               current coordinates.
        """
        # If a pspace was given, need to extract the dict from there, because
        # the steps below will lead to additional paramspace dimensions
        if pspace is not None:
            plot_cfg = recursive_update(copy.deepcopy(pspace._dict), plot_cfg)
            # FIXME internal API usage

        # Now have the plot config
        # Identify those keys that specify which universes to loop over
        try:
            unis = plot_cfg.pop('universes')
        
        except KeyError as err:
            raise ValueError("Missing required keyword-argument `universes` "
                             "in plot configuration!") from err

        # Get the parameter space, as it might be needed for certain values of
        # the `universes` argument
        self._psp = copy.deepcopy(self.psgrp.pspace)

        # If there was no parameter space available in the first place, only
        # the default point is available, which should be handled differently
        if self._psp.num_dims == 0 or self.psgrp.only_default_data_present:
            if unis not in ['all', 'single', 'first', 'random', 'any']:
                raise ValueError("Could not select a universe for plotting "
                                 "because the associated parameter space has "
                                 "no dimensions available or only data for "
                                 "the default point was available in {}. "
                                 "For these cases, the only valid values for "
                                 "the `universes` argument are: 'all', "
                                 "'single', 'first', 'random', or 'any'."
                                 "".format(self.psgrp.logstr))

            # Set a flag to carry information to _prepare_plot_func_args
            self._without_pspace = True

            # Distinguish cases where plot_cfg was given and those were not
            if pspace is not None:
                # There was a recursive update step; return the plot config
                # as parameter space
                return dict(), ParamSpace(plot_cfg)

            # else: Only need to return the plot configuration
            return plot_cfg, None

        # Parse it such that it is a valid subspace selector
        if isinstance(unis, str):
            if unis in ['all']:
                # is equivalent to an empty specifier -> empty dict
                unis = dict()

            elif unis in ['single', 'first', 'random', 'any']:
                # Find the state number from the universes available in the
                # parameter space group. Then retrieve the coordinates from
                # the corresponding parameter space state map

                # Create a list of available universe IDs
                uni_ids = [int(_id) for _id in self.psgrp.keys()]

                # Select the first or a random ID
                if unis in ['single', 'first']:
                    uni_id = min(uni_ids)
                else:
                    uni_id = np.random.choice(uni_ids)

                # Now retrieve the point from the (full) state map
                smap = self._psp.state_map
                point = smap.where(smap == uni_id, drop=True)
                # NOTE Universe IDs are unique, so that's ok.

                # And find its coordinates
                unis = {k: c.item() for k, c in point.coords.items()}

            else:
                raise ValueError("Invalid value for `universes` argument. Got "
                                 "'{}', but expected one of: 'all', 'single', "
                                 "'first', 'random', or 'any'.".format(unis))

        elif not isinstance(unis, dict):
            raise TypeError("Need parameter `universes` to be either a "
                            "string or a dictionary of subspace selectors, "
                            "but got: {} {}.".format(type(unis), unis))

        # else: was a dict, can be used as a subspace selector

        # Ensure that no invalid dimension names were selected
        for pdim_name in unis.keys():
            if pdim_name not in self._psp.dims.keys():
                raise ValueError("No parameter dimension '{}' was available "
                                 "in the parameter space associated with {}! "
                                 "Available parameter dimensions: {}"
                                 "".format(pdim_name, self.psgrp.logstr,
                                           ", ".join([n for n
                                                      in self._psp.dims])))

        # Copy parameter dimension objects for each coordinate
        # As the parameter space with the coordinates has a different
        # hierarchy than psp, the ordering has to be manually adjusted to be
        # the same as in the original psp.
        coords = dict()
        for dim_num, (name, pdim) in enumerate(self._psp.dims.items()):
            # Need to use a copy, as it will need to be changed
            _pdim = copy.deepcopy(pdim)

            # Make sure the name is the same as in the parameter space
            _pdim._name = name
            # FIXME internal API usage

            # Adjust the order to put them in front in the parameter space, but
            # keep the ordering the same as in `psp`.
            # NOTE The actual _order attribute does no longer play a role, as
            #      the psp.dims are already sorted according to those values.
            #      We just need to generate an offset to those parameter dims
            #      that might be defined in the plot configuration ...
            _pdim._order = -100000000 + dim_num
            # FIXME internal API usage
            
            # Now store it in the dict
            coords[name] = _pdim

        # Add these as a new key to the plot configuration
        if '_coords' in plot_cfg:
            raise ValueError("The given plot configuration may _not_ contain "
                             "the key '_coords' on the top-most level!")

        plot_cfg['_coords'] = coords

        # Convert the whole dict to a parameter space, the "multi plot config"
        mpc = ParamSpace(plot_cfg)

        # Activate only a certain subspace of the multi-plot configuration;
        # this determines which values will be iterated over.
        mpc.activate_subspace(**unis)

        # Now, also need the regular parameter space (i.e. without additional
        # plot configuration coordinates) to use in _prepare_plot_func_args.
        # Need to apply the universe selection to that as well
        self._psp.activate_subspace(**unis)
        self._psp_active_smap_cache = self._psp.active_state_map

        # Only return the configuration as a parameter space; all is included
        # in there now.
        return {}, mpc

    def _prepare_plot_func_args(self, *args, _coords: dict=None,
                                **kwargs) -> Tuple[tuple, dict]:
        """Prepares the arguments for the plot function and implements the
        special arguments required for ParamSpaceGroup-like data: selection of
        a single universe from the given coordinates.
        
        Args:
            *args: Passed along to parent method
            _coords (dict): The current coordinate descriptor which is then
                used to retrieve a certain point in parameter space from the
                state map attribute.
            **kwargs: Passed along to parent method
        
        Returns:
            tuple: (args, kwargs) for the plot function
        """
        # Need to distinguish between cases with or without pspace given. The
        # aim is to retrieve a Universe ID to use for selection.
        if self._without_pspace:
            # Only the default universe is available, always having ID 0.
            uni_id = 0

        else:
            # This is a parameter sweep.
            # Given the coordinates, retrieve the data for a single universe
            # from the state map. As _coords is created by the _prepare_cfg
            # method, it will unambiguously selects a universe ID.
            uni_id = int(self._psp_active_smap_cache.sel(**_coords))

        # Select the corresponding universe from the ParamSpaceGroup
        uni = self.psgrp[uni_id]
        log.note("Using data of:        %s", uni.logstr)
        
        # Let the parent function, implemented in ExternalPlotCreator, do its
        # thing. This will return the (args, kwargs) tuple and will also take
        # care of data transformation using the DAG framework, for which some
        # behaviour is specialized for selection from the passed `uni` using
        # additional helper methods; see below.
        return super()._prepare_plot_func_args(*args, uni=uni, **kwargs)

    def _get_dag_params(self, *, uni: ParamSpaceStateGroup, **cfg
                        ) -> Tuple[dict, dict]:
        """Makes the selected universe available and adjusts DAG parameters
        such that selections can be based on that universe.
        """
        dag_params, plot_kwargs = super()._get_dag_params(**cfg)

        # Extend the DAG parameters such that they perform a base_transform
        # to get to the selected universe and subsequently use that as the
        # selection base.
        uni_path = PATH_JOIN_CHAR.join([self.PSGRP_PATH, uni.name])
        base_transform = [dict(getitem=[DAGTag('dm'), uni_path], tag='uni',
                               file_cache=dict(read=False, write=False))]
        dag_params['init'] = dict(**dag_params['init'],
                                  base_transform=base_transform,
                                  select_base='uni')

        return dag_params, plot_kwargs



"""Definition of an OrderedDict-subclass that maintains the order key values
rather than the insertion order. It can be specialized to use a specific
comparison function for ordering keys.
"""

import collections
from typing import Callable

# Imports needed for replicating OrderedDict behaviour
import sys as _sys
from operator import eq as _eq
from collections import _proxy, _Link, _OrderedDictKeysView
from collections import _OrderedDictValuesView, _OrderedDictItemsView
from collections.abc import MutableMapping
from reprlib import recursive_repr as _recursive_repr

# -----------------------------------------------------------------------------

class KeyOrderedDict(dict):
    """This dict maintains the order of keys not by their insertion but by
    their value. It is a re-implementation of collections.OrderedDict, because
    subclassing that class is really difficult.

    Ordering is maintained by adjusting those methods of OrderedDict that
    take care of building and maintaining the doubly-linked list that provides
    the ordering. See OrderedDict for details, e.g. the weak-referencing.

    In effect, this relates only to __setitem__; all other methods rely on this
    to add elements to the mapping.

    For comparison, the ``key`` callable given at initialisation can be used to
    perform a operation on keys, the result of which is used in comparison.
    If this is not given, the ``DEFAULT_KEY_COMPARATOR`` class variable is
    used; note that this needs to be a binary function, where the first
    argument is equivalent to ``self`` and the second is the actual key to
    perform the unary operation on.
    """

    # The default key comparator
    DEFAULT_KEY_COMPARATOR = lambda _, k: k

    def __init__(self, *args, key: Callable=None, **kwds):
        """Initialize a KeyOrderedDict, which maintains key ordering. If no
        custom ordering function is given, orders by simple "smaller than"
        comparison.
        
        Apart from that, the interface is the same as for regular dictionaries.
        
        Args:
            *args: a single sequence of (key, value) pairs to insert
            key (Callable, optional): The callable used to 
            **kwds: Passed on to ``update`` method
        
        Raises:
            TypeError: on len(args) > 1
        """
        if len(args) > 1:
            raise TypeError("expected at most 1 arguments, got {}"
                            "".format(len(args)))

        # Set the key comparison function, passing through if nothing is set
        self._key = key if key is not None else self.DEFAULT_KEY_COMPARATOR

        # Set up internally used attributes
        try:
            self.__root

        except AttributeError:
            self.__hardroot = _Link()
            self.__root = root = _proxy(self.__hardroot)
            root.prev = root.next = root
            self.__map = {}

        # Populate the dict
        self.__update(*args, **kwds)

    # .........................................................................
    # Custom implementations that ascertain key ordering

    def _key_comp_lt(self, k1, k2) -> bool:
        """The key comparator. Returns true for k1 < k2.
        
        Before comparison, the unary ``self._key`` method is invoked on both
        keys.
        
        Args:
            k1: lhs of comparison
            k2: rhs of comparison
        
        Returns:
            bool: k1 < k2
        """
        # Create the actual keys to compare
        try:
            _k1, _k2 = self._key(k1), self._key(k2)
        
        except Exception as exc:
            raise ValueError("Could not apply key transformation method on "
                             "one or both of the keys '{}' and '{}'!"
                             "".format(k1, k2)
                             ) from exc
        
        # Compare
        try:
            return _k1 < _k2

        except Exception as exc:
            raise ValueError("Failed comparing '{}' (type {}, from key '{}') "
                             "to '{}' (type {}, from key '{}')! Keys of {} "
                             "need to be comparable after the key "
                             "transformation function {} was applied to them."
                             "".format(_k1, type(_k1), k1,
                                       _k2, type(_k2), k2,
                                       self.__class__.__name__,
                                       self._key.__name__)
                             ) from exc

    def __setitem__(self, key, value, *,
                    dict_setitem=dict.__setitem__, proxy=_proxy, Link=_Link):
        """Set the item with the provided key to the given value, maintaining
        the ordering specified by NEW_KEY_GT_NEXT_KEY

        If the key is not available, this takes care of finding the right place
        to insert the new link in the doubly-linked list.
        """
        if key not in self:
            # Create a new link in the inherited mapping
            self.__map[key] = link = Link()

            # Get the sentinel element and the one at the end
            root = self.__root
            last = root.prev

            # Find the link in the doubly-linked list after which the new
            # element is to be inserted, i.e.: the first element that compares
            # true when invoking the key comparator

            # Start with root element. For empty maps, this is already the
            # element after which the link can be inserted.
            element = root
            
            # For populated maps, need to iterate over the linked list
            if last is not root:
                # Start at root and iterate until again reaching root
                while element.next is not root:
                    # Check if the next element's key would compare strictly
                    # greater than the key to be inserted.
                    # NOTE root.key is not available!
                    if self._key_comp_lt(key, element.next.key):
                        # Found it.
                        break
                    # Did not find it. Go to the next one
                    element = element.next

            # element is now the element _after_ which to insert the new link
            link.prev, link.next, link.key = element, element.next, key

            # Update neighbouring links such that the new link is in between
            link.prev.next = link
            link.next.prev = proxy(link)  # need be a weak link

        # Key is now available
        # Set the value of the item using the method implemented by dict
        dict_setitem(self, key, value)

    def insert(self, key, value, *, hint_after=None, hint_before=None):
        """Inserts a (key, value) pair using hint information to speed up
        key search.
        """
        raise NotImplementedError

    # .........................................................................
    # Remaining dict interface, oriented at collections.OrderedDict

    def __delitem__(self, key, dict_delitem=dict.__delitem__):
        """kod.__delitem__(y) <==> del kod[y]
        
        Deleting an existing item uses self.__map to find the link which gets
        removed by updating the links in the predecessor and successor nodes.
        """
        dict_delitem(self, key)
        link = self.__map.pop(key)
        link_prev = link.prev
        link_next = link.next
        link_prev.next = link_next
        link_next.prev = link_prev
        link.prev = None
        link.next = None

    def __iter__(self):
        """kod.__iter__() <==> iter(kod)

        Traverse the linked list in order.
        """
        root = self.__root
        curr = root.next
        while curr is not root:
            yield curr.key
            curr = curr.next

    def __reversed__(self):
        """kod.__reversed__() <==> reversed(kod)
        
        Traverse the linked list in reverse order.
        """
        root = self.__root
        curr = root.prev
        while curr is not root:
            yield curr.key
            curr = curr.prev

    def clear(self):
        """Remove all items"""
        root = self.__root
        root.prev = root.next = root
        self.__map.clear()
        dict.clear(self)


    def __sizeof__(self):
        """Get the size of this object"""
        sizeof = _sys.getsizeof

        # number of links including root
        n = len(self) + 1

        # instance dictionary
        size = sizeof(self.__dict__)

        # internal dict and inherited dict
        size += sizeof(self.__map) * 2

        # link and proxy objects
        size += sizeof(self.__hardroot) * n
        size += sizeof(self.__root) * n

        # key comparator
        size += sizeof(self._key)

        return size

    update = __update = MutableMapping.update

    def keys(self):
        """D.keys() -> a set-like object providing a view on D's keys"""
        return _OrderedDictKeysView(self)

    def items(self):
        """D.items() -> a set-like object providing a view on D's items"""
        return _OrderedDictItemsView(self)

    def values(self):
        """D.values() -> an object providing a view on D's values"""
        return _OrderedDictValuesView(self)

    __ne__ = MutableMapping.__ne__

    __marker = object()

    def pop(self, key, default=__marker):
        """kod.pop(k[,d]) -> v, remove specified key and return the
        corresponding value.  If key is not found, default is returned if
        given, otherwise KeyError is raised.
        """
        if key in self:
            result = self[key]
            del self[key]
            return result
        if default is self.__marker:
            raise KeyError(key)
        return default


    def setdefault(self, key, default=None):
        """kod.setdefault(k[,d]) -> kod.get(k,d), also set kod[k]=d if k not
        in kod
        """
        if key in self:
            return self[key]
        self[key] = default
        return default

    @_recursive_repr()
    def __repr__(self):
        """kod.__repr__() <==> repr(kod)"""
        if not self:
            return '%s()' % (self.__class__.__name__,)
        return '%s(%r)' % (self.__class__.__name__, list(self.items()))

    def __reduce__(self):
        """Return state information for pickling"""
        inst_dict = vars(self).copy()
        for k in vars(KeyOrderedDict()):
            inst_dict.pop(k, None)
        return self.__class__, (), inst_dict or None, None, iter(self.items())

    def copy(self):
        """kod.copy() -> a shallow copy of kod, maintaining the key comparator"""
        return self.__class__(self, key=self._key)

    @classmethod
    def fromkeys(cls, iterable, value=None, *, key: Callable=None):
        """KOD.fromkeys(S[, v]) -> New key-ordered dictionary with keys from S.
        If not specified, the value defaults to None.
        """
        self = cls(key=key)
        for key in iterable:
            self[key] = value
        return self

    def __eq__(self, other):
        """kod.__eq__(y) <==> kod==y.
        Comparison to another KeyOrderedDict or OrderedDict is order-sensitive
        while comparison to a regular mapping is order-insensitive.
        """
        if isinstance(other, (KeyOrderedDict, collections.OrderedDict)):
            return dict.__eq__(self, other) and all(map(_eq, self, other))
        return dict.__eq__(self, other)


# -----------------------------------------------------------------------------
# Specializations

class IntOrderedDict(KeyOrderedDict):
    """A :py:class:`~dantro.utils.ordereddict.KeyOrderedDict` specialization
    that assumes keys to be castable to integer and using the comparison of
    the resulting integer values for maintaining the order
    """
    DEFAULT_KEY_COMPARATOR = lambda _, k: int(k)



"""Holds basic plot functions for use with ExternalPlotCreator"""

from ...data_mngr import DataManager
from ..pcr_ext import is_plot_func
from .mpl_setup import *


# Local constants
log = logging.getLogger(__name__)


# -----------------------------------------------------------------------------

@is_plot_func(creator_name='external', use_helper=False)
def lineplot(dm: DataManager, *, out_path: str, y: str, x: str=None,
             fmt: str=None, save_kwargs: dict=None, **plot_kwargs):
    """Performs a simple lineplot.
    
    Args:
        dm (DataManager): The data manager from which to retrieve the data
        out_path (str): Where to store the plot to
        y (str): The path to get to the y-data in the data manager
        x (str, optional): The path to get to the x-data in the data manager
        save_kwargs (dict, optional): kwargs to the plt.savefig function
        **plt_kwargs: Passed on to plt.plot
    """
    # Get the data
    x_data = dm[x] if x else None
    y_data = dm[y]

    # Assemble the arguments
    args = [x_data, y_data] if x_data is not None else [y_data]
    if fmt:
        args.append(fmt)

    # Call the plot function
    plt.plot(*args, **plot_kwargs)

    # Save and close figure
    plt.savefig(out_path, **(save_kwargs if save_kwargs else {}))
    plt.close()



"""This sub-module implements the basic mixin classes that are required
in the dantro.base module"""

import sys
import logging
import warnings

from ..abc import AbstractDataProxy, PATH_JOIN_CHAR
from ..tools import TTY_COLS

# Local constants
log = logging.getLogger(__name__)


class UnexpectedTypeWarning(UserWarning):
    """Given when there was an unexpected type passed to a data container."""
    pass


# -----------------------------------------------------------------------------

class AttrsMixin:
    """This Mixin class supplies the `attrs` property getter and setter and
    the private `_attrs` attribute.

    Hereby, the setter function will initialize a BaseDataAttrs-derived object
    and store it as an attribute.
    This relays the checking of the correct attribute format to the actual
    BaseDataAttrs-derived class.

    For changing the class that is used for the attributes, an overwrite of the
    _ATTRS_CLS class variable suffices.
    """
    # The class attribute that the attributes will be stored to
    _attrs = None

    # Define the class to use for storing attributes
    _ATTRS_CLS = None

    @property
    def attrs(self):
        """The container attributes."""
        return self._attrs

    @attrs.setter
    def attrs(self, new_attrs):
        """Setter method for the container `attrs` attribute."""
        if self._ATTRS_CLS is None:
            raise ValueError("Need to declare the class variable _ATTRS_CLS "
                             "in order to use the AttrsMixin!")

        # Perform the initialisation
        self._attrs = self._ATTRS_CLS(name='attrs', attrs=new_attrs)


class SizeOfMixin:
    """Provides the __sizeof__ magic method and attempts to take into account
    the size of the attributes.
    """

    def __sizeof__(self) -> int:
        """Returns the size of the data (in bytes) stored in this container's
        data and its attributes.

        Note that this value is approximate. It is computed by calling the
        ``sys.getsizeof`` function on the data, the attributes, the name and
        some caching attributes that each dantro data tree class contains.
        Importantly, this is not a recursive algorithm.

        Also, derived classes might implement further attributes that are not
        taken into account either. To be more precise in a subclass, create a
        specific __sizeof__ method and invoke this parent method additionally.

        For more information, see the documentation of ``sys.getsizeof``:

            https://docs.python.org/3/library/sys.html#sys.getsizeof
        """
        nbytes =  sys.getsizeof(self._data)
        nbytes += sys.getsizeof(self._attrs)
        nbytes += sys.getsizeof(self._name)
        nbytes += sys.getsizeof(self._logstr)

        return nbytes


class LockDataMixin:
    """This Mixin class provides a flag for marking the data of a group or
    container as locked.
    """
    # Whether the data is regarded as locked. Note name-mangling here.
    __locked = False

    @property
    def locked(self) -> bool:
        """Whether this object is locked"""
        return self.__locked

    def lock(self):
        """Locks the data of this object"""
        self.__locked = True
        self._lock_hook()
    
    def unlock(self):
        """Unlocks the data of this object"""
        self.__locked = False
        self._unlock_hook()

    def raise_if_locked(self, *, prefix: str=None):
        """Raises an exception if this object is locked; does nothing otherwise
        """
        if self.locked:
            raise RuntimeError("{}Cannot modify {} because it was already "
                               "marked locked."
                               "".format(prefix + " " if prefix else "",
                                         self.logstr))

    def _lock_hook(self):
        """Invoked upon locking."""
        pass
    
    def _unlock_hook(self):
        """Invoked upon unlocking."""
        pass
    

class CollectionMixin:
    """This Mixin class implements the methods needed for being a Collection.
    
    It relays all calls forward to the data attribute.
    """

    def __contains__(self, key) -> bool:
        """Whether the given key is contained in the items."""
        return bool(key in self.data)

    def __len__(self) -> int:
        """The number of items."""
        return len(self.data)

    def __iter__(self):
        """Iterates over the items."""
        return iter(self.data)


class ItemAccessMixin:
    """This Mixin class implements the methods needed for getting, setting,
    and deleting items. It relays all calls forward to the data attribute, but
    if given a list (passed down from above), it extracts it
    """

    def __getitem__(self, key):
        """Returns an item."""
        return self.data[self._item_access_convert_list_key(key)]

    def __setitem__(self, key, val):
        """Sets an item."""
        self.data[self._item_access_convert_list_key(key)] = val

    def __delitem__(self, key):
        """Deletes an item"""
        del self.data[self._item_access_convert_list_key(key)]

    def _item_access_convert_list_key(self, key):
        """If given something that is not a list, just return that key"""
        if isinstance(key, list):
            if len(key) > 1:
                return tuple(key)
            return key[0]
        return key


class MappingAccessMixin(ItemAccessMixin, CollectionMixin):
    """Supplies all methods that are needed for Mapping access.

    All calls are relayed to the data attribute.
    """

    def keys(self):
        """Returns an iterator over the attribute names."""
        return self.data.keys()

    def values(self):
        """Returns an iterator over the attribute values."""
        return self.data.values()

    def items(self):
        """Returns an iterator over the (keys, values) tuple of the attributes."""
        return self.data.items()

    def get(self, key, default=None):
        """Return the value at `key`, or `default` if `key` is not available."""
        return self.data.get(key, default)


class CheckDataMixin:
    """This mixin class extends a BaseDataContainer-derived class to check the
    provided data before storing it in the container.
    
    It implements a general _check_data method, overwriting the placeholder 
    method in the BaseDataContainer, and can be controlled via class variables.

    .. note::

        This is not suitable for checking containers that are added to an
        object of a BaseDataGroup-derived class!
    
    Attributes:
        DATA_ALLOW_PROXY (bool): Whether to allow _all_ proxy types, i.e.
            classes derived from AbstractDataProxy
        DATA_EXPECTED_TYPES (tuple, None): Which types to allow. If None, all
            types are allowed.
        DATA_UNEXPECTED_ACTION (str): The action to take when an unexpected
            type was supplied. Can be: raise, warn, ignore
    """

    # Specify expected data types for this container class
    DATA_EXPECTED_TYPES = None       # as tuple or None (allow all)
    DATA_ALLOW_PROXY = False         # to check for AbstractDataProxy
    DATA_UNEXPECTED_ACTION = 'warn'  # Can be: raise, warn, ignore

    def _check_data(self, data) -> None:
        """A general method to check the received data for its type
        
        Args:
            data: The data to check
        
        Raises:
            TypeError: If the type was unexpected and the action was 'raise'
            ValueError: Illegal value for DATA_UNEXPECTED_ACTION class variable
        
        Returns:
            None
        """
        if self.DATA_EXPECTED_TYPES is None:
            # All types allowed
            return

        # Compile tuple of allowed types
        expected_types = self.DATA_EXPECTED_TYPES

        if self.DATA_ALLOW_PROXY:
            expected_types += (AbstractDataProxy,)

        # Check for expected types
        if isinstance(data, expected_types):
            return

        # else: was not of the expected type
        # Create a base message
        msg = ("Unexpected type {} for data passed to {}! "
               "Expected types are: {}.".format(type(data), self.logstr,
                                                expected_types))

        # Handle according to the specified action
        if self.DATA_UNEXPECTED_ACTION == 'raise':
            raise TypeError(msg)

        elif self.DATA_UNEXPECTED_ACTION == 'warn':
            warnings.warn(msg + "\nInitialization will work, but be informed "
                          "that there might be errors at runtime.",
                          UnexpectedTypeWarning)
        
        elif self.DATA_UNEXPECTED_ACTION == 'ignore':
            log.debug(msg + " Ignoring ...")

        else:
            raise ValueError("Illegal value '{}' for class variable "
                             "DATA_UNEXPECTED_ACTION of {}. "
                             "Allowed values are: raise, warn, ignore"
                             "".format(self.DATA_UNEXPECTED_ACTION,
                                       self.classname))
